#include <../../nrnconf.h>
// solver CVode stub to allow cvode as dll for mswindows version.

#include <InterViews/resource.h>
#include "classreg.h"
#include "nrnoc2iv.h"
#include "datapath.h"
#if USECVODE
#include "cvodeobj.h"
#include "netcvode.h"
#else
class Cvode;
#endif

extern "C" {
void cvode_fadvance(double);
void cvode_finitialize();
boolean at_time(double);

extern double dt, t;
extern int nrn_cvode_;
extern void nrn_random_play();
extern int cvode_active_;
extern int nrn_use_daspk_;

Cvode* cvode_instance;
NetCvode* net_cvode_instance;
void deliver_net_events();
void nrn_deliver_events(double);
void clear_event_queue();
void init_net_events();
void nrn_record_init();
void nrn_play_init();
void fixed_record_continuous();
void fixed_play_continuous();
void nrn_solver_prepare();
}


// for fixed step fadvance
void deliver_net_events() {
	if (net_cvode_instance) {
		net_cvode_instance->deliver_net_events();
	}
}

// handle events during finitialize()
void nrn_deliver_events(double til) {
	double tsav = t;
	if (net_cvode_instance) {
		net_cvode_instance->deliver_events(til);
	}
	t = tsav;
}

void clear_event_queue() {
	if (net_cvode_instance) {
		net_cvode_instance->clear_events();
	}
}

void init_net_events() {
	if (net_cvode_instance) {
		net_cvode_instance->init_events();
	}
}

void nrn_record_init() {
	if (net_cvode_instance) {
		net_cvode_instance->record_init();
	}
}

void nrn_play_init() {
	if (net_cvode_instance) {
		net_cvode_instance->play_init();
	}
}

void fixed_play_continuous() {
	if (net_cvode_instance) {
		net_cvode_instance->fixed_play_continuous();
	}
}

void fixed_record_continuous() {
	if (net_cvode_instance) {
		net_cvode_instance->fixed_record_continuous();
	}
}

void nrn_solver_prepare() {
	if (net_cvode_instance) {
		net_cvode_instance->solver_prepare();
	}
}

void cvode_fadvance(double tstop) { // tstop = -1 means single step
#if USECVODE
	int err;
	if (net_cvode_instance) {
		nrn_random_play();
		err = net_cvode_instance->solve(tstop);
		if (err != 0) {
			printf("err=%d\n", err);
			hoc_execerror("variable step integrator error", 0);
		}
	}
#endif
}

void cvode_finitialize(){
#if USECVODE
	if (net_cvode_instance) {
		net_cvode_instance->re_init();
	}
#endif
}

boolean at_time(double te) {
#if USECVODE
	if (cvode_active_ && cvode_instance) {
		return cvode_instance->at_time(te);
	}
#endif
	double x = te - 1e-11;
	if (x <= t && x > (t - dt)) {
		return 1;
	}
	return 0;
}

/* no longer use a dll because we no longer care that neuron.exe as a zip
file does not fit onto a single floppy */

#if 0 && defined(WIN32)
extern "C" {

#include "cvodemlh.h"
#include "cvdense.h"
#include "cvdiag.h"

extern char* neuron_home;

typedef void (*CVF)(void* cvode_mem);
static void (*cvode_free)(void* cvode_mem);
void CVodeFree(void* cvode_mem) {
	(*cvode_free)(cvode_mem);
}

typedef void* (*CVMAL)(integer N, RhsFn f, real t0, N_Vector y0, int lmm, int iter,
		  int itol, real *reltol, void *abstol, void *f_data,
		  FILE *errfp, bool optIn, long int iopt[], real ropt[],
		  void *machEnv);
static void* (*cvode_malloc)(integer N, RhsFn f, real t0, N_Vector y0, int lmm, int iter,
		  int itol, real *reltol, void *abstol, void *f_data,
		  FILE *errfp, bool optIn, long int iopt[], real ropt[],
		  void *machEnv);
void *CVodeMalloc(integer N, RhsFn f, real t0, N_Vector y0, int lmm, int iter,
		  int itol, real *reltol, void *abstol, void *f_data,
		  FILE *errfp, bool optIn, long int iopt[], real ropt[],
		  void *machEnv)
{
	return (*cvode_malloc)(N, f, t0, y0, lmm, iter,
		itol, reltol, abstol, f_data,
		errfp, optIn, iopt, ropt,
		machEnv);
}

typedef int (*CVCV)(void *cvode_mem, real tout, N_Vector yout, real *t, int itask);
static int (*cvode_cvode)(void *cvode_mem, real tout, N_Vector yout, real *t, int itask);
int CVode(void *cvode_mem, real tout, N_Vector yout, real *t, int itask) {
	return (*cvode_cvode)(cvode_mem, tout, yout, t, itask);
}

typedef void (*CVDEN)(void *cvode_mem, CVDenseJacFn djac, void *jac_data);
static void (*cvode_dense)(void *cvode_mem, CVDenseJacFn djac, void *jac_data);
void CVDense(void *cvode_mem, CVDenseJacFn djac, void *jac_data) {
	(*cvode_dense)(cvode_mem, djac, jac_data);
}

typedef void (*CVDIA)(void* cvode_mem);
static void (*cvode_diag)(void* cvode_mem);
void CVDiag(void* cvode_mem) {
	(*cvode_diag)(cvode_mem);
}

typedef void (*NVF)(N_Vector x);
static void (*cvode_nvfree)(N_Vector x);
void N_VFree(N_Vector x){
	(*cvode_nvfree)(x);
}

typedef N_Vector (*NVN)(integer n, void *machEnv);
static N_Vector (*cvode_nvnew)(integer n, void *machEnv);
N_Vector N_VNew(integer n, void *machEnv) {
	return (*cvode_nvnew)(n, machEnv);
}

} // end extern c

#include "../windll/dll.h"
boolean load_cvode() {
	struct DLL* dll;
   char buf[256];
   sprintf(buf, "%s\\lib\\occvode.dll", neuron_home);
   dll = dll_load(buf);
   if (dll) {
   	Pfri reg = (Pfri)dll_lookup(dll, "_CVode_reg");
      if (reg) {
      	(*reg)();
      }
	cvode_free = (CVF)dll_lookup(dll, "_CVodeFree");
	cvode_malloc = (CVMAL)dll_lookup(dll, "_CVodeMalloc");
	cvode_cvode = (CVCV)dll_lookup(dll, "_CVode");
	cvode_dense = (CVDEN)dll_lookup(dll, "_CVDense");
	cvode_diag = (CVDIA)dll_lookup(dll, "_CVDiag");
	cvode_nvfree = (NVF)dll_lookup(dll, "_N_VFree");
	cvode_nvnew = (NVN)dll_lookup(dll, "_N_VNew");
	return true;
   }else{
   	printf("No CVode class.\n");
	return false;
   }
}
#endif
