#ifndef netcon_h
#define netcon_h

#undef check
#if MAC
#define NetCon nrniv_Dinfo
#endif

#include <InterViews/observe.h>
#include <OS/list.h>
#include "ivocvect.h"

#if 1
#define STATISTICS(arg) ++arg
#else
#define STATISTICS /**/
#endif

class PreSyn;
class PlayRecord;
class Cvode;
class TQueue;
class TQItem;
class NetCvode;
class SelfEventPool;
class HocEventPool;
class HocCommand;
class NetConSaveWeightTable;
class NetConSaveIndexTable;
class PreSynSaveIndexTable;

#define DiscreteEventType 0
#define TstopEventType 1
#define NetConType 2
#define SelfEventType 3
#define PreSynType 4
#define HocEventType 5
#define PlayRecordEventType 6
// the above will in turn steer to proper PlayRecord type

class DiscreteEvent : public Observer {
public:
	DiscreteEvent();
	virtual ~DiscreteEvent();
	virtual void send(double deliverytime, NetCvode*);
	virtual void deliver(double t, NetCvode*);
	virtual void pr(const char*, double t, NetCvode*);
	virtual void disconnect(Observable*) {};

	virtual DiscreteEvent* savestate_save();
	virtual void savestate_restore(double deliverytime, NetCvode*);
	virtual void savestate_write(FILE*);
	static DiscreteEvent* savestate_read(FILE*);

	// actions performed over each item in the event queue.
	virtual void frecord_init(TQItem*) {};

	static int discretevent_send_;
	static int discretevent_deliver_;
};

class NetCon : public DiscreteEvent {
public:
	NetCon();
	NetCon(PreSyn* src, Object* target);
	virtual ~NetCon();
	virtual void send(double sendtime, NetCvode*);
	virtual void deliver(double, NetCvode*);
	virtual void pr(const char*, double t, NetCvode*);

	virtual DiscreteEvent* savestate_save();
	static DiscreteEvent* savestate_read(FILE*);

	double delay_;
	boolean active_;
	PreSyn* src_;
	Point_process* target_;
	int cnt_;
	double* weight_;
	Object* obj_;
	void chksrc();
	void chktar();
	virtual void disconnect(Observable*);

	static int netcon_send_active_;
	static int netcon_send_inactive_;
	static int netcon_deliver_;
};
class NetConSave : public DiscreteEvent {
public:
	NetConSave(NetCon*);
	virtual ~NetConSave();
	virtual void savestate_restore(double deliverytime, NetCvode*);
	virtual void savestate_write(FILE*);
	NetCon* netcon_;	

	static void invalid();
	static NetCon* weight2netcon(double*);
	static NetCon* index2netcon(long);
private:
	static NetConSaveWeightTable* wtable_;
	static NetConSaveIndexTable* idxtable_;
};

class SelfEvent : public DiscreteEvent {
public:
	SelfEvent();
	virtual ~SelfEvent();
	virtual void deliver(double, NetCvode*);
	virtual void pr(const char*, double t, NetCvode*);
	void sefree();
	void clear(){} // called by sepool_->free_all

	virtual DiscreteEvent* savestate_save();
	virtual void savestate_restore(double deliverytime, NetCvode*);
	virtual void savestate_write(FILE*);
	static DiscreteEvent* savestate_read(FILE*);

	Point_process* target_;
	double* weight_;
	double flag_;
	void** movable_; // actually a TQItem**

	static SelfEvent* alloc();
	static void reclaim();
	static int selfevent_send_;
	static int selfevent_move_;
	static int selfevent_deliver_;
private:
	static SelfEventPool* sepool_;
};

declarePtrList(NetConPList, NetCon)

class PreSyn : public DiscreteEvent {
public:
	PreSyn(double* src, Object* osrc, Section* ssrc = nil);
	virtual ~PreSyn();
	virtual void send(double sendtime, NetCvode*);
	virtual void deliver(double, NetCvode*);
	virtual void pr(const char*, double t, NetCvode*);

	virtual DiscreteEvent* savestate_save();
	static DiscreteEvent* savestate_read(FILE*);

	void check();
	double* thvar_;
	Object* osrc_;
	Section* ssrc_;
	double threshold_;
	double after_spike_; // minimum time before another threshold crossing can occur.
	boolean flag_;
	NetConPList dil_;
	hoc_Item* hi_;
	long hi_index_; // for SaveState read and write
	void update(Observable*);
	int use_min_delay_;
	double delay_;
	void disconnect(Observable*);
	void record_stmt(const char*);
	void record(IvocVect*, IvocVect* idvec = nil, int rec_id = 0);
	void record(double t);
	void init();
	IvocVect* tvec_;
	IvocVect* idvec_;
	int rec_id_;
	HocCommand* stmt_;

	void condition(Cvode*);
	double new_threshold_event(Cvode*);
	void condition_abandon_statistics(Cvode*);
	double varold_, told_;
	TQItem* qthresh_;

	static int presyn_condition_init_above_;
	static int presyn_condition_qthresh_;
	static int presyn_condition_abandon_;
	static int presyn_condition_eq_abandon_;
	static int presyn_condition_abandon_init_above_;
	static int presyn_condition_abandon_init_below_;
	static int presyn_condition_abandon_above_;
	static int presyn_condition_abandon_below_;
	static int presyn_send_mindelay_;
	static int presyn_send_direct_;
	static int presyn_deliver_qthresh_;
	static int presyn_deliver_netcon_;
	static int presyn_deliver_direct_;
	static int presyn_deliver_ncsend_;
};
class PreSynSave : public DiscreteEvent {
public:
	PreSynSave(PreSyn*);
	virtual ~PreSynSave();
	virtual void savestate_restore(double deliverytime, NetCvode*);
	virtual void savestate_write(FILE*);
	PreSyn* presyn_;	
	boolean have_qthresh_;

	static void invalid();
	static PreSyn* hindx2presyn(long);
private:
	static PreSynSaveIndexTable* idxtable_;
};


class HocEvent : public DiscreteEvent {
public:
	HocEvent();
	virtual ~HocEvent();
	virtual void pr(const char*, double t, NetCvode*);
	static HocEvent* alloc(const char* stmt);
	void hefree();
	void clear(); // called by hepool_->free_all
	virtual void deliver(double, NetCvode*);
	static void reclaim();

	virtual DiscreteEvent* savestate_save();
	virtual void savestate_restore(double deliverytime, NetCvode*);
	virtual void savestate_write(FILE*);
	static DiscreteEvent* savestate_read(FILE*);
	
	static int hocevent_send_;
	static int hocevent_deliver_;
private:
	HocCommand* stmt_;
	static HocEvent* next_del_;
	static HocEventPool* hepool_;
};

class TstopEvent : public DiscreteEvent {
public:
	TstopEvent();
	virtual ~TstopEvent();
	virtual void deliver(double t, NetCvode*);
	virtual void pr(const char*, double t, NetCvode*);

	virtual DiscreteEvent* savestate_save();
	virtual void savestate_restore(double deliverytime, NetCvode*);
	virtual void savestate_write(FILE*);
	static DiscreteEvent* savestate_read(FILE*);
};

#endif
