#include <../../nrnconf.h>
#undef check
#include <string.h>
#include <stdlib.h>
#include <OS/list.h>
#include "nrnoc2iv.h"
#include "classreg.h"
#include "vgatedch.h"

#if defined(__MWERKS__) && !defined(_MSC_VER)
#include <extras.h>
#define strdup _strdup
#endif

declarePtrList(ChannelList, ChannelType)
implementPtrList(ChannelList, ChannelType)

static ChannelList* channels;

typedef void(*Pf5)(int, Node**, double**, Datum**, int);
extern "C" {
void register_mech(char**, void(*)(Prop*), Pf5, Pf5, Pf5, Pf5, int, int);
void hoc_register_cvode(int, int(*)(int),
	int(*)(int, double**, double**, double*, Datum*, double*, int),
	int(*)(int, Node**, double**, Datum**, int),
	int(*)(int, Node**, double**, Datum**, int)
);
int nrn_get_mechtype(const char*);
void ion_reg(const char*, double);
Prop* need_memb(Symbol*);
void nrn_promote(Prop*, int, int);
}

static void check(const char* s) {   
        if (hoc_lookup(s)) {
                hoc_execerror(s, "already exists");
        }
}

static void nrn_alloc(Prop* prop) {
	ChannelType* c = channels->item(prop->type);
	c->alloc(prop);
}

static void nrn_init(int n, Node** nodes, double** data, Datum** pdata, int type) {
	ChannelType* c = channels->item(type);
	c->init(n, nodes, data, pdata);
}

static void nrn_cur(int n, Node** nodes, double** data, Datum** pdata, int type) {
	ChannelType* c = channels->item(type);
	c->cur(n, nodes, data, pdata);
}

static void nrn_jacob(int n, Node** nodes, double** data, Datum** pdata, int type) {
	ChannelType* c = channels->item(type);
	c->jacob(n, nodes, data, pdata);
}

static void nrn_state(int n, Node** nodes, double** data, Datum** pdata, int type) {
	ChannelType* c = channels->item(type);
	c->state(n, nodes, data, pdata);
}

static int ode_count(int type){
	ChannelType* c = channels->item(type);
	return c->count();
}
static int ode_map(int ieq, double** pv, double** pvdot,
    double* p, Datum* pd, double* atol, int type){
	ChannelType* c = channels->item(type);
	c->map(ieq, pv, pvdot, p, pd, atol);
	return 0;
}
static int ode_spec(int n, Node** nd, double** p, Datum** pd, int type){
	ChannelType* c = channels->item(type);
	c->spec(n, nd, p, pd);
	return 0;
}
static int ode_matsol(int n, Node** nd, double** p, Datum** pd, int type){
	ChannelType* c = channels->item(type);
	c->matsol(n, nd, p, pd);
	return 0;
}
// hoc interface to VGateTransRate
static double E1(void* v) {return ((VGateTransRate*)v)->E1(*getarg(1));}
static double E2(void* v) {return ((VGateTransRate*)v)->E2(*getarg(1));}
static double alpha(void* v) {return ((VGateTransRate*)v)->alpha(*getarg(1));}
static double beta(void* v) {return ((VGateTransRate*)v)->beta(*getarg(1));}
static double tau(void* v) {return ((VGateTransRate*)v)->tau(*getarg(1));}
static double inf(void* v) {return ((VGateTransRate*)v)->inf(*getarg(1));}
static double m1(void* v) {
	VGateTransRate* r = (VGateTransRate*)v;
	if (ifarg(1)) {
		r->m1_ = *getarg(1);
	}
	return r->m1_;
}
static double b1(void* v) {
	VGateTransRate* r = (VGateTransRate*)v;
	if (ifarg(1)) {
		r->b1_ = *getarg(1);
	}
	return r->b1_;
}
static double m2(void* v) {
	VGateTransRate* r = (VGateTransRate*)v;
	if (ifarg(1)) {
		r->m2_ = *getarg(1);
	}
	return r->m2_;
}
static double b2(void* v) {
	VGateTransRate* r = (VGateTransRate*)v;
	if (ifarg(1)) {
		r->b2_ = *getarg(1);
	}
	return r->b2_;
}
static double set(void* v) {
	VGateTransRate* r = (VGateTransRate*)v;
	r->set(*getarg(1), *getarg(2), *getarg(3), *getarg(4));
	return 0.;
}

static Member_func r_members[] = {
	"E1", E1,
	"E2", E2,
	"alpha", alpha,
	"beta", beta,
	"tau", tau,
	"inf", inf,
	"m1", m1,
	"b1", b1,
	"m2", m2,
	"b2", b2,
	"set", set,
	0, 0
};

static void* r_cons(Object* o) {
	VGateTransRate* r = new VGateTransRate();
	r->obj_ = o; // but not reffed here
	if (ifarg(4)) {
		set(r);
	}
	return r;
}

static void r_destruct(void* v) {
	VGateTransRate* r = (VGateTransRate*)v;
	delete r;
}

void VGateTransRate_reg() {
	class2oc("VGateTransRate", r_cons, r_destruct, r_members);
}

// hoc interface to HHChannelType2
static Object** rate(void* v) {
	HHChannelType2* c = (HHChannelType2*)v;
	int i = (int)chkarg(1, 0, 1);
	Object* o = c->rate(i)->obj_;
	return hoc_temp_objptr(o);
}

static double power(void* v) {
	return 1.;
}

static Member_func hh_members[] = {
	"power", power,
	0, 0
};

static Member_ret_obj_func hh_omembers[] = {
	"rate", rate,
	0, 0
};

static void* hh_cons(Object* o) {
	hoc_obj_ref(o); // so never destroyed
	char* suffix = gargstr(1);
	check(suffix);
	char* ion = gargstr(2);
	Object* t1 = *hoc_objgetarg(4);
	Object* t2 = *hoc_objgetarg(7);
	check_obj_type(t1, "VGateTransRate");
	check_obj_type(t2, "VGateTransRate");
	hoc_obj_ref(t1); // never destroyed
	hoc_obj_ref(t2); // never destroyed
	HHChannelType2* c = new HHChannelType2(suffix, ion,
		gargstr(3), (VGateTransRate*)t1->u.this_pointer, (int)(*getarg(5)),
		gargstr(6), (VGateTransRate*)t2->u.this_pointer, (int)(*getarg(8))
	);
	return c;
}
static void hh_destruct(void*) {
	assert(0);
}

void HHChannelType2_reg() {
	class2oc("HHChannelType2", hh_cons, hh_destruct, hh_members, nil,
		hh_omembers);
}

// param is gbar, g, m, h --- if change then change numbers below
static char* m_hh2[] = {"0", "hhna", "gbar", 0, "g", 0, "m", "h", 0, 0 };
// gbar=0 g=1 m=2 h=3

ChannelType::ChannelType() {}
ChannelType::~ChannelType() {}
void ChannelType::add_channel(char** m) {
	ChannelType* c = (ChannelType*)this;
	register_mech(m, nrn_alloc, nrn_cur, nrn_jacob, nrn_state, nrn_init, -1, 1);
	int type = nrn_get_mechtype(m[1]);
	hoc_register_cvode(type, ode_count, ode_map, ode_spec, ode_matsol);
	if (!channels) {
		channels = new ChannelList(50);
	}
	while(channels->count() < type) {
		channels->append(nil);
	}
	channels->append(c);
}

HHStateType::HHStateType(const char* name, VGateTransRate* r) {
	r_ = r;
	name_ = new char[strlen(name)+1];
	strcpy(name_, name);
}
HHStateType::~HHStateType() {
	delete [] name_;
}
HHConductanceType2::HHConductanceType2(int p1, int p2) {
	pow1 = p1; pow2 = p2;
}
HHConductanceType2::~HHConductanceType2() {
}

HHChannelType2::HHChannelType2(const char* suffix, const char* ion,
  const char* s1, VGateTransRate* r1, int pow1,
  const char* s2, VGateTransRate* r2, int pow2
  ) {
	int i;
	ion_reg(ion, -10000.);
	char buf[100];
	sprintf(buf, "%s_ion", ion);
	ion_sym_ = hoc_lookup(buf);
	if (!ion_sym_) {
		hoc_execerror(buf, " is not an ion mechanism");
	}
	check(suffix);
	m_hh2[1] = (char*)suffix;
	sprintf(buf, "gbar_%s", suffix); m_hh2[2] = strdup(buf);
	sprintf(buf, "g_%s", suffix); m_hh2[4] = strdup(buf);
	sprintf(buf, "%s_%s", s1, suffix); m_hh2[6] = strdup(buf);
	sprintf(buf, "%s_%s", s2, suffix); m_hh2[7] = strdup(buf);
	add_channel(m_hh2);
	for (i=2; i < 8; ++i) if (m_hh2[i]) { free(m_hh2[i]); }
	s_[0] = new HHStateType(s1, r1);
	s_[1] = new HHStateType(s2, r2);
	c2_ = new HHConductanceType2(pow1, pow2);
}

HHChannelType2::~HHChannelType2(){}

void HHChannelType2::alloc(Prop* prop) {
	prop->param_size = 6;
	prop->param = (double*)ecalloc(6, sizeof(double));
	Datum* pp = prop->dparam = (Datum*)hoc_Ecalloc(3, sizeof(Datum)); hoc_malchk();
	Prop* prop_ion = need_memb(ion_sym_);
	nrn_promote(prop_ion, 0, 1);
	pp[0].pval = prop_ion->param + 0; //ena
	pp[1].pval = prop_ion->param + 3; //ina
	pp[2].pval = prop_ion->param + 4; //dinadv
}

void HHChannelType2::init(int n, Node** nd, double** pp, Datum**) {
	int i;
	for (i=0; i < n; ++i) {
		double v = nd[i]->v;
		pp[i][2] = s_[0]->r()->inf(v);
		pp[i][3] = s_[1]->r()->inf(v);
	}
}

void HHChannelType2::state(int n, Node** nd, double** pp, Datum**) {
	int i;
	for (i=0; i < n; ++i) {
		double v = nd[i]->v;
		pp[i][2] = s_[0]->integrate(v, pp[i][2], dt);
		pp[i][3] = s_[1]->integrate(v, pp[i][3], dt);
	}
}

void HHChannelType2::cur(int n, Node** nd, double** pp, Datum** ppd) {
	int i;
	for (i=0; i < n; ++i) {
		double g = pp[i][1] = c2_->g(pp[i][0], pp[i][2], pp[i][3]);
		double ena = *ppd[i][0].pval;
		double ina = g * (nd[i]->v - ena);
		NODERHS(nd[i]) -= ina;

		*ppd[i][1].pval += ina; // not happy about this
		*ppd[i][2].pval += g; // why bother unless plotting current
	}
}

void HHChannelType2::jacob(int n, Node** nd, double** pp, Datum** ppd) {
	int i;
	for (i=0; i < n; ++i) {
		NODED(nd[i]) += pp[i][1];
	}
}

int HHChannelType2::count() { return 2; }

void HHChannelType2::map(int ieq, double** pv, double** pvdot,
    double* p, Datum* pd, double* atol){
	int i;
	cvode_ieq_ = ieq;
	for (i=0; i < 2; ++i) {
		pv[i] = p + 2 + i;
		pvdot[i] = p + 4 + i;
	}
}

void HHChannelType2::spec(int n, Node** nd, double** p, Datum** pd){
	int i;
	for (i=0; i < n; ++i) {
		double v = nd[i]->v;
		p[i][4] = s_[0]->dstate(v, p[i][2]);
		p[i][5] = s_[1]->dstate(v, p[i][3]);
	}
}

void HHChannelType2::matsol(int n, Node** nd, double** p, Datum** pd){
	int i;
	for (i=0; i < n; ++i) {
		double v = nd[i]->v;
		p[i][4] = s_[0]->matsol(p[i][4], v, dt);
		p[i][5] = s_[1]->matsol(p[i][5], v, dt);
	}
}

