#include <../../nrnconf.h>
/* /local/src/master/nrn/src/nrnoc/capac.c,v 1.6 1998/11/25 19:14:28 hines Exp */

#include	"section.h"
#include	"membdef.h"
#include	"membfunc.h"

static double cj; 
static char *mechanism[] = { "0", "capacitance", "cm",0, "i_cap", 0,0 };
static cap_alloc(), cap_init();

capac_reg_() {
	/* all methods deal with capacitance in special ways */
	register_mech(mechanism, cap_alloc, (Pfri)0, (Pfri)0, (Pfri)0, cap_init, -1, 1);
}

#define cm  vdata[i][0]
#define i_cap  vdata[i][1]
#define vm  vnode[i]->v

/*
cj is analogous to 1/dt for cvode and daspk
for fixed step second order it is 2/dt and
for pure implicit fixed step it is 1/dt
*/
void nrn_set_cj(x) double x; { /* x is 1/ms */
	cj = x;
}

double nrn_get_cj() { return cj; }

void nrn_cap_jacob(count, vnode, vdata )
/*ARGSUSED*/
	int count;
	Node *vnode[];
	double *vdata[];
{
	int i;
	double cfac = .001*cj;

	for (i=0; i < count; ++i) {
		NODED(vnode[i]) += cfac*cm;
	}
}

static cap_init(count, vnode, vdata, vpdata, type )
/*ARGSUSED*/
	int count, type;
	Node **vnode;
	double **vdata;
	Datum **vpdata;
{
	int i;
	for (i=0; i < count; ++i) {
		i_cap = 0;
	}
}

nrn_capacity_current(count, vnode, vdata)
	int count;
	Node **vnode;
	double **vdata;
{
	int i;
	double cfac = .001*cj;
	/* since rhs is dvm for a full or half implicit step */
	/* (nrn_update_2d() replaces dvi by dvi-dvx) */
	/* no need to distinguish secondorder */
	for (i=0; i < count; ++i) {
		i_cap = cfac*cm*NODERHS(vnode[i]);
	}
}

#if CVODE
nrn_mul_capacity(count, vnode, vdata)
	int count;
	Node **vnode;
	double **vdata;
{
	int i;
	double cfac = .001*cj;
	for (i=0; i < count; ++i) {
		NODERHS(vnode[i]) *= cfac*cm;
	}
}
nrn_div_capacity(count, vnode, vdata)
	int count;
	Node **vnode;
	double **vdata;
{
	int i;
	for (i=0; i < count; ++i) {
		i_cap = NODERHS(vnode[i]);
		NODERHS(vnode[i]) /= 1.e-3*cm;
	}
}
#endif /*CVODE*/

/* the rest can be constructed automattically from the above info*/

static cap_alloc(p)
	Prop *p;
{
	double *pd;
#define nparm 2
	pd = (double *)ecalloc(nparm, sizeof(double));
	pd[0] = DEF_cm;	/*default capacitance/cm^2*/
	p->param = pd;
	p->param_size = nparm;
}

