#include <../../nrnconf.h>
#include "bbsconf.h"
#include <InterViews/resource.h>
#include "classreg.h"
#include "oc2iv.h"
#include "ivocvect.h"
#include "hoclist.h"
#include "bbs.h"
#include "bbsimpl.h"
#include "ivocvect.h"
#include "parse.h"

extern "C" {
	extern int vector_arg_px(int, double**);
	Symbol* hoc_which_template(Symbol*);
	void bbs_done();
}

class OcBBS : public BBS , public Resource {
public:
	OcBBS(int nhost_request);
	virtual ~OcBBS();
public:
	double retval_;
	int userid_;
	int next_local_;
};

OcBBS::OcBBS(int n) : BBS(n) {
	next_local_ = 0;
}

OcBBS::~OcBBS() {
}

static boolean posting_ = false;
static void pack_help(int, OcBBS*);
static void unpack_help(int, OcBBS*);
static int submit_help(OcBBS*);
static char* key_help();

void bbs_done() {
#if USEBBS
	Symbol* sym = hoc_lookup("ParallelContext");
	sym = hoc_which_template(sym);
	hoc_Item* q, *ql;
	ql = sym->u.ctemplate->olist;
	q = ql->next;
	if (q != ql) {
		Object* ob = OBJ(q);
		OcBBS* bbs = (OcBBS*)ob->u.this_pointer;
		if (bbs->is_master()) {bbs->done();}
	}
#endif
}

static int submit_help(OcBBS* bbs) {
	int id, i, firstarg, style;
	posting_ = true;
	bbs->pkbegin();
	i = 1;
	if (hoc_is_double_arg(i)) {
		bbs->pkint((id = (int)chkarg(i++, 0, 1e7)));
	}else{
		bbs->pkint((id = --bbs->next_local_));
	}
	if (ifarg(i+1)) {
#if 1
		int argtypes = 0;
		int ii = 1;
		if (hoc_is_str_arg(i)) {
			style = 1;
			bbs->pkint(style); // "fname", arg1, ... style
		}else{
			style = 2;
			bbs->pkint(style); // [object],"fname", arg1, ... style
			Object* ob = *hoc_objgetarg(i++);
			bbs->pkstr(ob->ctemplate->sym->name);
			bbs->pkint(ob->index);
//printf("ob=%s\n", hoc_object_name(ob));
		}
		bbs->pkstr(gargstr(i++));
		firstarg = i;
		for (; ifarg(i); ++i) { // first is least significant
			if (hoc_is_double_arg(i)) {
				argtypes += 1*ii;
			}else if (hoc_is_str_arg(i)) {
				argtypes += 2*ii;
			}else{ // must be a Vector
				argtypes += 3*ii;
			}
			ii *= 4;
		}
//printf("submit style %d %s argtypes=%o\n", style, gargstr(firstarg-1), argtypes);
		bbs->pkint(argtypes);
		pack_help(firstarg, bbs);
#endif
	}else{
		bbs->pkint(0); // hoc statement style
		bbs->pkstr(gargstr(i));
	}
	posting_ = false;
	return id;
}

static double submit(void *v) {
	int id;
	OcBBS* bbs = (OcBBS*)v;
	id = submit_help(bbs);
	bbs->submit(id);
	return double(id);
}
	
static double context(void *v) {
	OcBBS* bbs = (OcBBS*)v;
	submit_help(bbs);
	bbs->context();
	return 1.;
}
	
static double working(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	int id;
	boolean b = bbs->working(id, bbs->retval_, bbs->userid_);
	if (b) {
		return double(id);
	}else{
		return 0.;
	}
}

static double retval(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	return bbs->retval_;
}

static double userid(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	return (double)bbs->userid_;
}

static double nhost(void* v) {
#if defined(HAVE_STL)
	OcBBS* bbs = (OcBBS*)v;
	return double(bbs->nhost());
#else
	return -1.;
#endif
}

static double worker(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	bbs->worker();
	return 0.;
}

static double done(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	bbs->done();
	return 0.;
}

static void pack_help(int i, OcBBS* bbs) {
	if (!posting_) {
		bbs->pkbegin();
		posting_ = true;
	}
	for (; ifarg(i); ++i) {
		if (hoc_is_double_arg(i)) {
			bbs->pkdouble(*getarg(i));
		}else if (hoc_is_str_arg(i)) {
			bbs->pkstr(gargstr(i));
		}else{
			int n; double* px;
			n = vector_arg_px(i, &px);
			bbs->pkint(n);
			bbs->pkvec(n, px);
		}
	}
}

static double pack(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	pack_help(1, bbs);
	return 0.;
}

static double post(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	pack_help(2, bbs);
	posting_ = false;
	if (hoc_is_str_arg(1)) {
		bbs->post(gargstr(1));
	}else{
		char key[50];
		sprintf(key, "%g", *getarg(1));
		bbs->post(key);
	}
	return 1.;
}

static void unpack_help(int i, OcBBS* bbs) {
	for (; ifarg(i); ++i) {
		if (hoc_is_pdouble_arg(i)) {
			*hoc_pgetarg(i) = bbs->upkdouble();
		}else if (hoc_is_str_arg(i)) {
			char* s = bbs->upkstr();
			char** ps = hoc_pgargstr(i);
			hoc_assign_str(ps, s);
			delete [] s;
		}else{
			Vect* vec = vector_arg(i);
			int n = bbs->upkint();
			vec->resize(n);
			bbs->upkvec(n, vec->vec());
		}
	}
}

static double unpack(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	unpack_help(1, bbs);
	return 1.;
}

static double upkscalar(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	return bbs->upkdouble();
}

static char** upkstr(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	char* s = bbs->upkstr();
	char** ps = hoc_pgargstr(1);
	hoc_assign_str(ps, s);
	delete [] s;
	return ps;
}

static Object** upkvec(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	Vect* vec;
	int n = bbs->upkint();
	if (ifarg(1)) {
		vec = vector_arg(1);
		vec->resize(n);
	}else{
		vec = new Vect(n);
	}
	bbs->upkvec(n, vec->vec());
	return vec->temp_objvar();
}

static char* key_help() {
	static char key[50];
	if (hoc_is_str_arg(1)) {
		return gargstr(1);
	}else{
		sprintf(key, "%g", *getarg(1));
		return key;
	}
}

static double take(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	bbs->take(key_help());
	unpack_help(2, bbs);
	return 1.;
}

static double look(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	if (bbs->look(key_help())) {
		unpack_help(2, bbs);
		return 1.;
	}
	return 0.;
}

static double look_take(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	if (bbs->look_take(key_help())) {
		unpack_help(2, bbs);
		return 1.;
	}
	return 0.;
}

static Member_func members[] = {
	"submit", submit,
	"working", working,
	"retval", retval,
	"userid", userid,
	"pack", pack,
	"post", post,
	"unpack", unpack,
	"upkscalar", upkscalar,
	"take", take,
	"look", look,
	"look_take", look_take,
	"runworker", worker,
	"done", done,
	"nhost", nhost,
	"context", context,
	0,0
};

static Member_ret_str_func retstr_members[] = {
	"upkstr", upkstr,
	0,0
};

static Member_ret_obj_func retobj_members[] = {
	"upkvec", upkvec,
	0,0
};

static void* cons(Object*) {
	// not clear at moment what is best way to handle nested context
	int i = -1;
	if (ifarg(1)) {
		i = int(chkarg(1, 0, 10000));
	}
	OcBBS* bbs = new OcBBS(i);
	bbs->ref();
	return bbs;
}

static void destruct(void* v) {
	OcBBS* bbs = (OcBBS*)v;
	bbs->unref();
}

void ParallelContext_reg() {
	class2oc("ParallelContext", cons, destruct, members, nil,
		retobj_members, retstr_members);
}

void BBSImpl::execute_helper() {
	char* s;
	int style = upkint();
	switch (style) {
	case 0:
		s = upkstr();
		hoc_obj_run(s, nil);
		delete [] s;
		break;
	case 1:
	case 2: {
#if 1
		int i, j;
		Symbol* fname;
		Object* ob = nil;
		char* sarg[20]; // upto 20 argument may be strings
		int ns = 0; // number of args that are strings
		int narg = 0; // total number of args
		if (style == 2) { // object first
			s = upkstr(); // template name
			i = upkint(); // object index
//printf("template |%s| index=%d\n", s, i);
			Symbol* sym = hoc_lookup(s);
			if (sym) {
				sym = hoc_which_template(sym);
			}
			if (!sym) {
				hoc_execerror(s, "is not a template");
			}
			hoc_Item* q, *ql;
			ql = sym->u.ctemplate->olist;
			ITERATE(q, ql) {
				ob = OBJ(q);
				if (ob->index == i) {
					break;
				}
				ob = nil;
			}
			if (!ob) {
fprintf(stderr, "%s[%d] is not an Object in this process\n", s, i);
hoc_execerror("ParallelContext execution error", 0);
			}
			delete [] s;
			s = upkstr();
			fname = hoc_table_lookup(s, sym->u.ctemplate->symtable);
		}else{
			s = upkstr();
			fname = hoc_lookup(s);
		}
//printf("execute helper style %d fname=%s obj=%s\n", style, fname->name, hoc_object_name(ob));
		if (!fname) {
fprintf(stderr, "%s not a function in %s\n", s, hoc_object_name(ob));
hoc_execerror("ParallelContext execution error", 0);
		}
		int argtypes = upkint(); // first is least signif
		for (j = argtypes; (i = j%4) != 0; j /= 4) {
			++narg;
			if (i == 1) {
				double x = upkdouble();
//printf("arg %d scalar %g\n", narg, x);
				hoc_pushx(x);
			}else if (i == 2) {
				sarg[ns] = upkstr();
//printf("arg %d string |%s|\n", narg, sarg[ns]);
				hoc_pushstr(sarg+ns);
				ns++;
			}else{
				int n;
				n = upkint();
				Vect* vec = new Vect(n);
//printf("arg %d vector size=%d\n", narg, n);
				upkvec(n, vec->vec());
				hoc_pushobj(vec->temp_objvar());
			}
		}
		hoc_ac_ = hoc_call_objfunc(fname, narg, ob);
		delete [] s;
		for (i=0; i < ns; ++i) {
			delete [] sarg[i];
		}
#endif
	    }
		break;
	}
}

void BBSImpl::return_args(int id) {
	// the message has been set up by the subclass
	// perhaps it would be better to do this directly
	// and avoid the meaningless create and delete.
	// but then they all would have to know this format
	int i;
	char* s;
//printf("BBSImpl::return_args(%d):\n", id);
	i = upkint(); // userid
	int style = upkint();
//printf("message userid=%d style=%d\n", i, style);
	switch (style) {
	case 0:
		s = upkstr(); // the statement
//printf("statement |%s|\n", s);
		delete [] s;
		break;
	case 2: // obj first
		s = upkstr(); // template name
		i = upkint();	// instance index
//printf("object %s[%d]\n", s, i);
		delete [] s;
		//fall through
	case 1:
		s = upkstr(); //fname
		i = upkint(); // arg manifest
//printf("fname=|%s| manifest=%o\n", s, i);
		delete [] s;
		break;
	}
	// now only args are left and ready to unpack.
}


